/* --8<--8<--8<--8<--
 *
 * Copyright (C) 2006 Smithsonian Astrophysical Observatory
 *
 * This file is part of luasup
 *
 * luasup is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * luasup is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the 
 *       Free Software Foundation, Inc. 
 *       51 Franklin Street, Fifth Floor
 *       Boston, MA  02110-1301, USA
 *
 * -->8-->8-->8-->8-- */

#include <stdio.h>
#include <string.h>

#include <lua.h>

#include <suplib/str.h>

#include "luasup.h"

/* -------------------------------------------------------------------------------*/
/*
  PROCEDURE
  slua_tbl_nelem - return the number of elements in a Lua table

  DESCRIPTION
  This routine counts the number of elements in a Lua table.

  RETURNS
  It returns the number of elements in the table

  AUTHOR
  Diab Jerius

*/

int
slua_tbl_nelem(
  lua_Object table 		/* the table to iterate over */
  )
{
  lua_Object index = LUA_NOOBJECT;
  int nelem = 0;

  while( LUA_NOOBJECT != slua_tbl_next( table, &index ) )
    nelem++;
  
  return nelem;
}

/* -------------------------------------------------------------------------------*/
/*
  PROCEDURE
  slua_tbl_next - an implementation of the Lua @code{next} function

  DESCRIPTION
  This routine is a wrapper around the Lua @code{next} function, for
  which there are no default bindings for C.  It is passed a table and
  an index.  The index should be set to @code{LUA_NOOBJECT} to start
  the scan in the table.  @code{slua_tbl_next} will update the index to
  point to the next element in the table, and will return the value
  in the table at that index.  Note that the @emph{address} of the
  index is passed to @code{slua_tbl_next}.

  If there are no more elements in the table, the returned value and
  the index are set to @code{LUA_NOOBJECT}.

  RETURNS
  It returns the value of the table at the next index, or
  @code{LUA_NOOBJECT} if there are no more elements.

  AUTHOR
  Diab Jerius

*/

lua_Object
slua_tbl_next(
  lua_Object table, 		/* the table to iterate over */
  lua_Object *index		/* A pointer to the last index used; set the
				   index's value to LUA_NOOBJECT on
				   the first pass */
  )
{
  lua_Object lua_next_func = lua_getglobal( "next" );

  lua_pushobject( table );
  if ( LUA_NOOBJECT == *index )
    lua_pushnil();
  else
    lua_pushobject( *index );
  lua_callfunction( lua_next_func );

  *index = lua_getparam( 1 );
  return lua_getparam( 2 );
}

/* -------------------------------------------------------------------------------*/
/*
  PROCEDURE
  slua_tbl_el_s - get a string element from a table

  DESCRIPTION
  retrieve the value of an element in a table. The value should be
  representable as a string, as a pointer to a dynamically allocated
  string is returned.  The calling routine is responsible for
  deallocating the string.

  RETURNS
  It returns zero if all went well, @samp{1} if the element didn't
  exist, and @samp{-1} if it was unable to allocate enough memory for
  the string. If the element doesn't exist, @code{*value}
  is unchanged.

  AUTHOR
  Diab Jerius */
int
slua_tbl_el_s(
  lua_Object table,		/* the table containing the data */
  const char *index,		/* the index of the element */
  char **value			/* the output value of the element */
  )
{
  lua_Object element;
  lua_pushobject( table );
  lua_pushstring( (char * )index );
  element = lua_gettable();

  if ( lua_isnil( element ) )
    return 1;
  
  *value = str_dup( lua_getstring( element ) );
    return NULL == *value ? -1 : 0;
}

/* -------------------------------------------------------------------------------*/
/*
  PROCEDURE
  slua_tbl_el_si - get a string element from a table

  DESCRIPTION
  retrieve the value of an element in a table. The value should be
  representable as a string, as a pointer to the
  string is returned.  This points to the Lua object;  it differs
  from @code{slua_tbl_el_s} in that it doesn't allocate any memory
  for the string.

  RETURNS
  It returns a pointer to the string upon succes, @code{NULL}
  if the element didn't exist.

  AUTHOR
  Diab Jerius
*/

char *
slua_tbl_el_si(
  lua_Object table,		/* the table containing the data */
  const char *index		/* the index of the element */
  )
{
  lua_Object element;
  lua_pushobject( table );
  lua_pushstring( (char *) index );
  element = lua_gettable();

  return lua_isnil( element ) ? NULL : lua_getstring( element );
}


/* -------------------------------------------------------------------------------*/
/*
  PROCEDURE
  slua_tbl_el_n - get a numeric element from a table

  DESCRIPTION
  retrieve the value of an element in a table. The value should be
  representable as a number.

  RETURNS
  It returns zero if all went well and @samp{1} if the element didn't
  exist.  If the element doesn't exist, @code{*value}
  is unchanged.

  AUTHOR
  Diab Jerius
*/
int
slua_tbl_el_n(
  lua_Object table,		/* the table containing the data */
  const char *index,		/* the index of the element */
  double *value			/* the output value of the element */
)
{
  lua_Object element;
  lua_pushobject( table );
  lua_pushstring( (char *) index );
  element = lua_gettable();

  if ( lua_isnil( element ) )
    return 1;

  *value = lua_getnumber( element );
  return 0;
}

/* -------------------------------------------------------------------------------*/
/*
  PROCEDURE
  slua_tbl_el_exists - determine existence of an element in a table

  DESCRIPTION
  This routine checks for the existence of an element in a table.

  RETURNS
  It returns zero if the element doesn't exist, non-zero otherwise.

  AUTHOR
  Diab Jerius
*/
int
slua_tbl_el_exists(
  lua_Object table,		/* the table containing the data */
  const char *index		/* the index of the element */
  )
{
  lua_Object element;
  lua_pushobject( table );
  lua_pushstring( (char *) index );
  element = lua_gettable();

  return lua_isnil( element ) ? 0 : 1;
}


/* -------------------------------------------------------------------------------*/
/*
  PROCEDURE
  slua_tbl_el_tbl - get a table element from a table

  DESCRIPTION
  retrieve the value of an element in a table. The value should be
  a table. The table handle is returned (no copy is made).

  RETURNS
  It returns the table handle, or @code{LUA_NOOBJECT} if the element
  didn't exist or wasn't a table.
  exist, and @samp{-1} if it was unable to allocate enough memory for
  the string. If the element doesn't exist, @code{*value}
  is unchanged.

  AUTHOR
  Diab Jerius
*/
lua_Object
slua_tbl_el_tbl(
  lua_Object table,		/* the table containing the data */
  const char *index		/* the index of the element */
  )
{
  lua_Object element;
  lua_pushobject( table );
  lua_pushstring( (char * )index );
  element = lua_gettable();

  if ( lua_isnil( element ) || ! lua_istable( element ))
    return LUA_NOOBJECT;
  
  return element;
}

/* -------------------------------------------------------------------------------*/
/*
  PROCEDURE
  slua_tbl_el_delete - delete an element from a table

  DESCRIPTION
  This routine sets the value of a specified element in a table to
  @samp{nil}, effectively deleting it.

  AUTHOR
  Diab Jerius
*/
void
slua_tbl_el_delete(
  lua_Object table,		/* the table containing the data */
  const char *index		/* the index of the element to delete */
)
{
  lua_pushobject( table );
  lua_pushstring( (char *) index );
  lua_pushnil( );
  lua_settable( );
}  

/* -------------------------------------------------------------------------------*/
/*
  PROCEDURE
  slua_tbl_cmp - compare two tables for equality

  DESCRIPTION
  This routine compares two tables for equality. it is a deep comparison.


  RETURNS
  @code{0} if the two tables are equivalent, @code{1} if they are not.

  AUTHOR
  Diab Jerius
*/
int
slua_tbl_cmp(
  lua_Object table1,		/* the first table */
  lua_Object table2		/* the second table */
)
{
  lua_Object index = LUA_NOOBJECT;
  lua_Object value1;
  lua_Object value2;

  /* iterate through table1, comparing values to table2 */
  while( LUA_NOOBJECT != ( value1 = slua_tbl_next( table1, &index ) ) )
  {
    SLuaArgType type = slua_type( value1 );

    lua_pushobject( table2 );
    lua_pushobject( index );
    value2 = lua_gettable();

    /* value2 will be nil if index isn't in the table */
    if ( type != slua_type( value2 ) )
      return 1;

    switch( type )
    {
    case SLua_STRING:
      if ( strcmp( lua_getstring( value1 ), lua_getstring( value2 ) ) )
	return 1;
      break;

    case SLua_NUMBER:
      if ( lua_getnumber( value1 ) != lua_getnumber( value2 ) )
	return 1;
      break;

    case SLua_TABLE:
      if ( slua_tbl_cmp( value1, value2 ) )
	return 1;
      break;
      
    case SLua_FUNCTION:
      if ( value1 != value2 )
	return 1;
      break;

    case SLua_CFUNCTION:
      if ( value1 != value2 )
	return 1;
      break;

    case SLua_USERDATA:
      if ( value1 != value2 )
	return 1;
      break;

    default:
      return 1;
      break;
    }
    
  }

  /* now, make sure there aren't any values in table2 that aren't in
     table 1 */
  index = LUA_NOOBJECT;
  while( LUA_NOOBJECT != ( value2 = slua_tbl_next( table2, &index ) ) )
  {
    lua_Object value2;

    lua_pushobject( table1 );
    lua_pushobject( index );

    value1 = lua_gettable();

    /* value2 will be nil if index isn't in the table */
    if ( lua_isnil( value1 ) )
      return 1;
  }


  return 0;
}  


/* -------------------------------------------------------------------------------*/
/*
  PROCEDURE
  slua_tbl_copy - make a shallow copy of a table

  DESCRIPTION
  This routine makes a shallow copy of a table. It doesn't not attempt
  to recursively create tables contained within the table.

  RETURNS
  It returns a table object upon success, or @code{LUA_NOOBJECT} upon error.

  AUTHOR
  Diab Jerius
*/
lua_Object
slua_tbl_copy(
  lua_Object table		/* the table to copy */
)
{
  lua_Object index = LUA_NOOBJECT;
  lua_Object value;
  lua_Object copy = lua_createtable();
  

  if (  LUA_NOOBJECT == copy )
    return copy;

  while( LUA_NOOBJECT != ( value = slua_tbl_next( table, &index ) ) )
  {
    lua_pushobject( copy );
    lua_pushobject( index );
    lua_pushobject( value );
    lua_settable( );
  }

  return copy;
}  


/* -------------------------------------------------------------------------------*/
/*
  PROCEDURE
  slua_tbl_el - get an element from a table

  DESCRIPTION
  This function retrieves the value of an element with index
  @code{index} in a table.  The caller must specify the type of
  argument to expect.  If the type is or'd with @code{SLua_OPTIONAL}
  the element need not be present.  If the type is @code{SLua_STRING}
  and is or'd with @code{SLua_COPY}, a copy of the string value is
  returned (the caller must handle deallocation).

  The value of the element is stored in the passed @code{dest} variable.

  Upon error, the passed function @code{errfunc} is called with an
  error message.  The message will be prefixed with the string in
  @code{err_pfx} (using @code{"%s: error message"}).  @code{errfunc}
  may be @code{NULL}, in which case it won't be called.

  RETURNS
  It returns one of the following values:
  @table @code

  @item 0

  everything went ok or the passed table
  was equal to @code{LUA_NOOBJECT}.

  @item SLua_ENOEL

  The element was optional, but not present.

  @item SLua_ENOENT
  The element was not in the table.

  @item SLua_ENOMEM
  The type was @code{SLua_STRING|SLua_COPY} and the string could not be
  duplicated.

  @item SLua_ENOTYPE
  The type is unknown.  
  
  @item SLua_ENOMATCH
  The requested type and the actual type do not agree.

  @end table

  AUTHOR
  Diab Jerius

*/
SLuaError
slua_tbl_el(
  SLua_ErrFunc errfunc,	/* the error function */
  lua_Object table,	/* the table */
  const char *err_pfx, 	/* a prefix for error messages */
  const char *index,	/* the index of the element to delete */
  SLuaArgType type,	/* the type of argument to expect (with or'd flags )*/
  void *dest		/* where to store the retrieved data */
  )
{
  lua_Object element;

  /* harmless if the table doesn't exist */
  if ( LUA_NOOBJECT == table )
    return SLua_ENOERR;

  /* get a handle to the element */
  lua_pushobject( table );
  lua_pushstring( (char * )index );
  element = lua_gettable();

  /* if the element doesn't exist, error out if it isn't optional,
     else return indicating it doesn't */
  if ( lua_isnil( element ) )
  {
    if ( SLua_OPTIONAL & type )
      return SLua_ENOEL;
    else
    {
      if ( errfunc )
	errfunc( "%s: table element `%s' not found", err_pfx, index );
      return SLua_ENOENT;
    }
  }

  /* it exists, handle it specially, depending upon what it is */

  switch( type & SLua_TYPES )
  {
    /* number */
  case SLua_NUMBER:
    if ( lua_isnumber( element ) )
    {
      *((double *) dest) = lua_getnumber( element );
      return SLua_ENOERR;
    }
    break;

    /* string */
  case SLua_STRING:
    if ( lua_isstring( element ) )
    {
      if ( SLua_COPY & type )
      {
	if ( NULL ==
	     ( *((char **) dest) = str_dup( lua_getstring( element ) )) )
	  return  SLua_ENOMEM;
	else
	  return SLua_ENOERR;
      }
      else
      {
	*((char **) dest) = lua_getstring( element );
	return SLua_ENOERR;
      }
    }
    break;

    /* table */
  case SLua_TABLE:
    if ( lua_istable( element ) )
    {
      *((lua_Object *)dest) = element;
      return SLua_ENOERR;
    }
    break;

  default:
    if ( errfunc )
      errfunc( "%s: unknown lua object type `%d'. client error"
	       " in call to `slua_tbl_el' in library 'luasup'\n",
	       err_pfx, type & SLua_TYPES );
    return SLua_ENOTYPE;
    break;
  }

  /* if we've made it this far, it means that the type did
     not match what was requested */
  if ( errfunc )
    errfunc( "%s: incorrect type for `%s'. expected a %s",
	     err_pfx, index, slua_type_name( type ) );
  return SLua_ENOMATCH;
}
